#ifndef XPDEFINES_H
#define XPDEFINES_H

#include <string>

const std::string defaultConfigFile     = "default";
const std::string defaultFont		= "FreeMonoBold.ttf";
const std::string defaultMapExt		= ".jpg";
const std::string defaultStarMap        = "BSC";
const std::string separator             = "/";

#endif
